/**
 * @(#)PolyLineGroup.java - used to store all lines max, min, PolyLine
 *
 * Created on August 1, 2000, 9:05 PM
 */


/**
 * PolyLineGroup is a container of PolyLines. The container is needed
 * for:
 *
 *  1) calculating the scaling factors since they have to be the same
 *     along all PolyLines in this group
 *  2) calculating the relative portions of the PolyLines for each
 *     value of the abscissae
 *
 * After the PolyLines got filled with values and the PolyLines are
 * added to the group renderGroup() is called. This functions
 * calculates the various windows, the scaling factors and the grid.
 *
 * @author  Raimar Falke <rf13@inf.tu-dresden.de>
 * @version 0.10
 */

/**
   ********************************************************************
 CivLog - Copyleft (C) 2000 - Wong Teck Meng
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
**********************************************************************
                                                                      */
package org.freeciv.civlog.util;

import org.freeciv.civlog.Debug;
import org.freeciv.civlog.ColorManager;

import java.awt.*;
import java.util.*;

public class PolyLineGroup extends Object
{
    private Vector lines=new Vector();

    /*
     * internalOutputWindow and finalOutputWindow are relative to
     * graphicsOutputWindow. So point(x,y) for internalOutputWindow is
     * the same as point(x,y) for finalOutputWindow is the same as the
     * point(x,y) for the surroundings component.
     */

    /*
     * This is the whole window which can be used to draw things. Is
     * given as an argument to the constructor.
     */
    Dimension graphicsOutputWindow;

    /*
     * All outside of internalOutputWindow is used for decoration and
     * is not available to indivual PolyLines.
     */
    Rectangle internalOutputWindow=new Rectangle();

    /*
     * Space outside of finalOutputWindow can be reserved by providing
     * numbers which are measured in pixel to the constructor.
     */
    Rectangle finalOutputWindow=new Rectangle();

    private final boolean debugRendering = false;
    private final boolean showInternalOutputWindow=false;

    private boolean isGridDash = false;
    private float scaleFactorX, scaleFactorY;
    boolean isRelativeGraph;

    float maxX, maxY, minX, minY;
    Graphics graphics;
    int spaceTop, spaceBottom, spaceLeft, spaceRight;

    float xGridLines[];
    String xGridLinesLabel[];
    int renderedXGridLines[];
    float finalXGridLineDistance;

    float yGridLines[];
    String yGridLinesLabel[];
    int renderedYGridLines[];
    float finalYGridLineDistance;

    String xLabel,yLabel;

    public PolyLineGroup (Graphics graphics, Dimension graphicsOutputWindow,
                          int spaceTop, int spaceBottom,
                          int spaceLeft, int spaceRight,
                          String xLabel, String yLabel,
                          boolean isRelativeGraph)
    {
        this.graphics=graphics;
        this.graphicsOutputWindow=graphicsOutputWindow;
        this.spaceTop=spaceTop;
        this.spaceBottom=spaceBottom;
        this.spaceLeft=spaceLeft;
        this.spaceRight=spaceRight;
        this.xLabel=xLabel;
        this.yLabel=yLabel;
        this.isRelativeGraph=isRelativeGraph;
    }

    public void addPolyLine(PolyLine line)
    {
        lines.addElement(line);
        line.setGroup(this);
    }

    public PolyLine getLine(int index)
    {
        return (PolyLine)lines.elementAt(index);
    }

    public void renderGroup()
    {
        trace("PolyLineGroup: graphics output window ["+
              graphicsOutputWindow.width+"x"+graphicsOutputWindow.height+"]");

        renderGroup0();
        renderGroup1();
        renderGroup2();
        renderGroup3();
        renderGroup4();
    }

    /*
     * Calculates the percentage of each PolyLine's y for all
     * points. The points must have the same x value.
     */
    public void renderGroup0()
    {
        if(!isRelativeGraph)
            return;

        int points=0;
        for (int i=0; i < lines.size(); i++)
        {
            PolyLine currentLine=(PolyLine)lines.elementAt(i);
            points=Math.max(points, currentLine.getNoOfPoints());
        }

        for(int point=0;point<points;point++)
        {
            float currentX=0.0f;
            boolean first=true;
            float sum=0.0f;
            float yValues[]=new float[lines.size()+1];

            for (int line=0; line < lines.size(); line++)
            {
                PolyLine currentLine=(PolyLine)lines.elementAt(line);

                float x=currentLine.getX(point);

                if(first)
                {
                    first=false;
                    currentX=x;
                }
                if(currentX!=x)
                {
                    throw new InternalError();
                }

                yValues[line+1]=currentLine.getY(point);
                sum+=yValues[line+1];
            }

            if (sum==0.0)
            {
                final float reciprocal=100.0f/(float)lines.size();

                for (int i=1;i<yValues.length;i++)
                {
                    yValues[i]=reciprocal;
                }
            }
            else
            {
                sum/=100.0f;
                for (int line=1;line<yValues.length;line++)
                {
                    yValues[line]/=sum;
                }
            }

            yValues[0]=100.0f;
            for (int line=0; line < lines.size(); line++)
            {
                PolyLine currentLine=(PolyLine)lines.elementAt(line);

                yValues[line+1]=yValues[line]-yValues[line+1];

                currentLine.setRelativeRange(point,yValues[line+1],
                                             yValues[line]);
            }
        }
    }

    /*
     * Calculate the minimum and maximum values for x and y for all
     * PolyLines in this group.
     */
    public void renderGroup1()
    {
        maxX=-Float.MAX_VALUE;
        maxY=-Float.MAX_VALUE;
        minX=Float.MAX_VALUE;
        minY=Float.MAX_VALUE;

        for (int i=0; i < lines.size(); i++)
        {
            PolyLine currentLine=(PolyLine)lines.elementAt(i);
            currentLine.startRendering();

            if(currentLine.maxX>maxX)
                maxX=currentLine.maxX;

            if(currentLine.maxY>maxY)
                maxY=currentLine.maxY;

            if(currentLine.minX<minX)
                minX=currentLine.minX;

            if(currentLine.minY<minY)
                minY=currentLine.minY;
        }
        if(isRelativeGraph)
        {
            minY=0.0f;
            maxY=100.0f;
        }
    }

    /*
     * Calculate the positions of the grid lines, and the internal
     * output window.
     */
    public void renderGroup2()
    {
        float resultX[]=getGridLineDistance(minX,maxX,
                                            graphicsOutputWindow.width);
        minX=resultX[0];
        maxX=resultX[1];
        finalXGridLineDistance=resultX[2];

        float resultY[]=getGridLineDistance(minY,maxY,
                                            graphicsOutputWindow.height);
        minY=resultY[0];
        maxY=resultY[1];
        finalYGridLineDistance=resultY[2];


        /*
         * Fill xGridLines and xGridLinesLabel.
         */
        int finalXGridLines=(int)((maxX-minX)/finalXGridLineDistance)+1;
        trace("PolyLineGroup: after adjusting x range there will be "+
              finalXGridLines+" grid lines");

        xGridLines=new float[finalXGridLines];
        xGridLinesLabel=new String[finalXGridLines];

        for(int i=0;i<finalXGridLines;i++)
        {
            xGridLines[i]=finalXGridLineDistance*i+minX;
            xGridLinesLabel[i]=""+xGridLines[i];
        }

        /*
         * Fill yGridLines and yGridLinesLabel.
         */
        int finalYGridLines=(int)((maxY-minY)/finalYGridLineDistance)+1;
        trace("PolyLineGroup: after adjusting y range there will be "+
              finalYGridLines+" grid lines");

        int maxWidth=0;
        FontMetrics fm=graphics.getFontMetrics();
        yGridLines=new float[finalYGridLines];
        yGridLinesLabel=new String[finalYGridLines];
        for(int i=0;i<finalYGridLines;i++)
        {
            yGridLines[i]=finalYGridLineDistance*i+minY;
            yGridLinesLabel[i]=""+yGridLines[i];
            maxWidth=Math.max(maxWidth,fm.stringWidth(yGridLinesLabel[i]));
        }

        /*
         * Set internalOutputWindow
         */

        // reserve space to put the yGridLinesLabel; the labels are
        // set 10 pixel from the left border
        internalOutputWindow.x=maxWidth+10;

        // reserve 2 font heights for the yLabel at the top
        internalOutputWindow.y=2*fm.getHeight();

        // reserver 1 pixel this let showInternalOutputWindow work
        internalOutputWindow.width=graphicsOutputWindow.width-
                                   internalOutputWindow.x-1;

        // at the bottom reserve 4 font heights for the two rows of
        // xGridLinesLabel, the xLabel and some space between these
        // three rows
        internalOutputWindow.height=graphicsOutputWindow.height-
                                    internalOutputWindow.y-
                                    4*fm.getHeight();

        trace("PolyLineGroup: internal output window ("+
              internalOutputWindow.x+","+internalOutputWindow.y+
              ") ["+internalOutputWindow.width+"x"+
              internalOutputWindow.height+"]");
    }

    /*
     * Calculate the final output window and the scaling
     * factors.
     */
    public void renderGroup3()
    {
        float width=maxX-minX;
        float height=maxY-minY;

        trace("PolyLineGroup: source rectangle ("+minX+","+minY+") ["+
              width+"x"+height+"]");

        /*
         * Calculate final output window by reserving requested
         * space. Narrow the internalOutputWindow by the given space
         * values.
         */
        finalOutputWindow.x=internalOutputWindow.x+spaceLeft;
        finalOutputWindow.width=(internalOutputWindow.width-
                                 spaceLeft-spaceRight);

        finalOutputWindow.y=internalOutputWindow.y+spaceTop;
        finalOutputWindow.height=(internalOutputWindow.height-
                                  spaceTop-spaceBottom);

        trace("PolyLineGroup: final output window ("+
              finalOutputWindow.x+","+
              finalOutputWindow.y+") ["+finalOutputWindow.width+"x"+
              finalOutputWindow.height+"]");

        scaleFactorX=(float)finalOutputWindow.width/width;
        scaleFactorY=(float)finalOutputWindow.height/height;

        /*
         * Give the data to all lines. They will render all their
         * points.
         */
        for (int i=0; i < lines.size(); i++)
        {
            PolyLine currentLine=(PolyLine)lines.elementAt(i);

            currentLine.finishRendering(scaleFactorX,scaleFactorY,
                                        finalOutputWindow);
        }
    }

    /*
     * Render the grid lines.
     */
    public void renderGroup4()
    {
        renderedXGridLines=new int[xGridLines.length];
        for(int i=0;i<renderedXGridLines.length;i++)
        {
            renderedXGridLines[i]=Math.round((xGridLines[i]-minX)*
                                             scaleFactorX)+finalOutputWindow.x;
        }

        renderedYGridLines=new int[yGridLines.length];
        for(int i=0;i<renderedYGridLines.length;i++)
        {
            renderedYGridLines[i]=finalOutputWindow.height-
                                  Math.round((yGridLines[i]-minY)*
                                             scaleFactorY)+finalOutputWindow.y;
        }
    }

    public float[] unrender(int x, int y)
    {
        float result[]=new float[2];

        x-=finalOutputWindow.x;
        y-=finalOutputWindow.y;

        result[0]=x/scaleFactorX+minX;
        result[1]=y/scaleFactorY+minY;

        return result;
    }

    /*
     * Calculate the positions of the grid lines.
    * Returns an array with three elements: the new min in index 0, 
    * max in [1] and distance in [2].
     */
    private float[] getGridLineDistance(float min, float max,
                                        int graphicsSpace)
    {
        trace ("PolyLineGroup: getGridLineDistance():");

        float result[]=new float[3];

        if(max==min)
        {
            result[0]=min;
            result[1]=min+1.0f;
            result[2]=1.0f;
        }
        else
        {
            /*
             * Get a rough idea where to put the grid lines
             */
            float exp=(float)(Math.log(max-min)/Math.log(10));
            exp=(float)Math.floor(exp)-1.0f;

            int minNumberOfGridLines=graphicsSpace/80;
            int maxNumberOfGridLines=graphicsSpace/33;

            trace("PolyLineGroup:  want between "+
                  minNumberOfGridLines+" and "+
                  maxNumberOfGridLines+" lines");

            /*
             * Start from a point where there are in any case enough
             * lines. Now increase the distance between lines and check if
             * the number of lines get into the valid range. The prefered
             * values are: 10**n 2.5*10**n and 5.0*10**n.
             */
            float interimHeight=max-min;
            for(int i=(int)exp;;i++)
            {
                float baseGridLineDistance=(float)Math.pow(10.0,i);
                float gridLineDistance;
                int lines;

                gridLineDistance=baseGridLineDistance*1.0f;
                lines=(int)(interimHeight/gridLineDistance);
                if(lines>=minNumberOfGridLines && lines<=maxNumberOfGridLines)
                {
                    result[2]=gridLineDistance;
                    break;
                }

                gridLineDistance=baseGridLineDistance*2.5f;
                lines=(int)(interimHeight/gridLineDistance);
                if(lines>=minNumberOfGridLines && lines<=maxNumberOfGridLines)
                {
                    result[2]=gridLineDistance;
                    break;
                }

                gridLineDistance=baseGridLineDistance*5.0f;
                lines=(int)(interimHeight/gridLineDistance);
                if(lines>=minNumberOfGridLines && lines<=maxNumberOfGridLines)
                {
                    result[2]=gridLineDistance;
                    break;
                }
                if(lines==0)
                    System.exit(0);
            }
            trace("PolyLineGroup: final distance between grid lines is "+
                  result[2]+"; there will be "+
                  (int)((max-min)/result[2])+" lines");

            /*
             * Adjust min and max
             */
            trace("PolyLineGroup: old range: "+min+" - "+max);
            result[0]=((float)(Math.floor(min/result[2]))*
                       result[2]);
            result[1]=((float)(Math.ceil(max/result[2]))*
                       result[2]);
            trace("PolyLineGroup: new range: "+
                  result[0]+" - "+result[1]);
        }
        return result;
    }

    public void drawGrid ()
    {
        Color gridColor = ColorManager.getGridColor();

        Graphics2D g2 = (Graphics2D) graphics;

        FontMetrics fm = g2.getFontMetrics();
        int height = fm.getHeight ();

        /** save stroke */
        BasicStroke oldStroke = (BasicStroke) g2.getStroke ();

        if ( isGridDash )
        {
            /** made grid draw using dashed line */
            float dash[] = {8.0f};
            g2.setStroke (new BasicStroke (oldStroke.getLineWidth () / 2,
                                           oldStroke.getEndCap (),
                                           oldStroke.getLineJoin (),
                                           8.0f,
                                           dash,
                                           oldStroke.getMiterLimit () ));
        }

        g2.setColor (gridColor);
        g2.drawRect (finalOutputWindow.x,
                     finalOutputWindow.y,
                     finalOutputWindow.width,
                     finalOutputWindow.height);

        for(int i=0; i<renderedXGridLines.length; i++)
        {
            g2.setColor (gridColor);
            g2.drawLine (renderedXGridLines[i],
                         finalOutputWindow.y,
                         renderedXGridLines[i],
                         finalOutputWindow.height+finalOutputWindow.y);

            int width=fm.stringWidth(xGridLinesLabel[i]);
            int offsetY;

            if(i%2 == 0)
                offsetY = 0;
            else
                offsetY =- height;

            g2.setColor (ColorManager.getXAxisNumbersColor());
            g2.drawString (xGridLinesLabel[i],
                           renderedXGridLines[i]-width/2,
                           graphicsOutputWindow.height-(height*3)/2+
                           offsetY);
        }

        for(int i=0;i<renderedYGridLines.length;i++)
        {
            g2.setColor(gridColor);
            g2.drawLine (finalOutputWindow.x,
                         renderedYGridLines[i],
                         finalOutputWindow.width+finalOutputWindow.x,
                         renderedYGridLines[i]);
            g2.setColor (ColorManager.getYAxisNumbersColor());
            g2.drawString(yGridLinesLabel[i], 5,
                          renderedYGridLines[i] + 3);
        }

        g2.setColor (ColorManager.getXAxisLabelColor());
        int width=fm.stringWidth(xLabel);
        g2.drawString(xLabel,
                      (finalOutputWindow.x+finalOutputWindow.width)/2-
                      width/2,
                      graphicsOutputWindow.height-height/2);

        g2.setColor (ColorManager.getYAxisLabelColor());
        g2.drawString(yLabel,
                      finalOutputWindow.x+20,
                      (height*3)/2); // we have 2*height space

        // reset back to old stroke
        g2.setStroke (oldStroke);

        if(showInternalOutputWindow)
        {
            graphics.setColor(Color.green);
            graphics.drawRect(internalOutputWindow.x,
                              internalOutputWindow.y,
                              internalOutputWindow.width,
                              internalOutputWindow.height);
        }
    }

    public void setShowDashGrid(boolean b)
    {
        isGridDash = b;
    }

    private void trace(String msg)
    {
        if(debugRendering)
            Debug.trace(msg);
    }
}
// End of PolyLineGroup.java
